/*
 * pagefault_coro2.c
 *
 * Use FLEX's page fault handling interface to do extra work during page faults
 * This one does the extra work outside the signal handler and thus is better
 *
 * This is not working because of:
 * 1) Mucking with the context argument in the signal handler cause the process
 *    to segfault somehow
 * 2) The context got from the signal handler cannot be used in setcontext().
 *    It also segfaults.
 */

#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <signal.h>
#include <flex.h>
#include <ucontext.h>
#include "coro.h"

/* Number of reads to perform */
#define N 100
#define MAX_WORKER_NR 10

io_context_t ioctx;
struct coroutine *co_scheduler;
struct coroutine *co_workers[MAX_WORKER_NR];
int nr_worker;

/* ucontext of the yielding function */
ucontext_t uctx_yield, uctx_main;

#define STACK_SIZE (64*1024)
#define STATE_RUNNABLE 0
#define STATE_BLOCKED 1
#define STATE_FINISHED 2

struct worker_context {
    /* scheduler stuff */
    int state;
    ucontext_t uctx;        /* where we should continue from after page fault is done */
    unsigned long faulting_address; /* index to match against the completion event */
    
    /* user stuff */
    int id;
    char *buffer;
    int length;
    char worker_stack[STACK_SIZE];
};

char scheduler_stack[STACK_SIZE];
char yield_function_stack[STACK_SIZE];

void pagefault_yield() {
    printf("1st from pagefault_yield()!\n");
    swapcontext(&uctx_yield, &uctx_main);
    
    while (1) {
        struct worker_context *wctx;
        /* Tell the scheduler to suspend this thread */
        co_call (co_scheduler, 0);
        /* Back fromt he scheduler, let the thread continue */
        wctx = (struct worker_context *)co_current->user;
        setcontext(&wctx->uctx);
    }
}

void pagefault_handler(int sig, struct siginfo *info, void *_ctx) {
    struct worker_context *wctx;
    static int count;
    ucontext_t *ctx = (ucontext_t *)_ctx;
    (void)sig;
    (void)info;

    printf("From pagefault_handler()\n");

    if (count++==0) {
        return;
    }
    /* Store return context and faulting address in worker context */
    wctx = (struct worker_context *)co_current->user;
    memcpy(&wctx->uctx, ctx, sizeof(ucontext_t));
    wctx->state = STATE_BLOCKED;
    wctx->faulting_address = (unsigned long) info->si_addr;

    /* Let signal handler finish and "return" to the pagefault_yield() */
    memcpy(ctx, &uctx_yield, sizeof(ucontext_t));
}

/* Worker coroutine that reads from a mmap'ed buffer */
void worker(void *_ctx) {
    struct worker_context *ctx;
    int i, off, s;

    /* Get context from main coroutine */
    ctx = (struct worker_context *)_ctx;

    for (i = 0; i < N; i++) {
        off = rand() % (ctx->length);
        s += ctx->buffer[off];
    }

    printf("Coroutine %d done\n", ctx->id);
    co_call(co_scheduler, (void *)(-1));
}

/* Fake scheduler coroutine */
void scheduler() {
    struct io_event events[1];
    struct timespec *wait;
    struct timespec zerowait;
    int i;

    zerowait.tv_sec = 0;
    zerowait.tv_nsec = 0;
    
    /* Keep polling for completion of the page fault */
    while (1) {
        int r;
        struct coroutine *next;
        struct worker_context *wctx;
        int alive_nr;
        printf(".");
        
        /* Do not blocking waiting for events when we have something to run */
        wait = &zerowait;

wait_for_events:
        /* Test whether some coroutines become runnable again */
        if (io_getevents(ioctx, 1, 1, events, wait) == 1) {
            unsigned long address = (unsigned long) events[0].data;
            int success = 0;
            for (i = 0; i < nr_worker; i++) {
                wctx = (struct worker_context *)co_workers[i]->user;
                if (wctx->state == STATE_BLOCKED && wctx->faulting_address == address) {
                    wctx->state = STATE_RUNNABLE;
                    success = 1;
                    break;
                }
            }
            if (!success) {
                printf("Get unmatched completion event: address=%lux\n", address);
                exit(-1);
            }
        }

        /* Scheduler a thread to run */
        next = 0;
        alive_nr = 0;
        for (i = 0; i < nr_worker; i++) {
            wctx = (struct worker_context *)co_workers[i]->user;
            if (wctx->state != STATE_FINISHED)
                alive_nr ++;
            if (wctx->state == STATE_RUNNABLE) {
                next = co_workers[i];
                break;
            }
        }
        if (alive_nr == 0) {
            printf("Finished, sheduler exiting\n");
            exit(0);
        }
        if (!next) {
            /* No thread runnable, then wait  */
            printf("No thread available to run, waiting\n");
            
            wait = 0;  /* block waiting for events */
            goto wait_for_events;
        }

        r = (int)co_call(next, next->user);
        switch(r) {
        case 0:             /* Page fault happens */
            printf("o");
            break;
        case -11:           /* The thread finishes */
            wctx->state = STATE_FINISHED;
            break;
        }
    }
}


int main(int argc, char **argv) {
    int i;
    struct stat st;
    struct sigaction sa;
     
    if (argc < 2) {
        printf("Pass filename as argument\n");
        exit(0);
    }
    
    /* Set up signal handler to receive page fault notification */
    sa.sa_sigaction = pagefault_handler;
    sa.sa_flags = SA_SIGINFO;

    if (sigaction(SIGPF, &sa, 0) < 0) {
        perror ("Error setting up signal handler");
        exit(-1);
    }

    /* Set up an io context and specify it to receive page faults events */
    if (io_setup(1024, &ioctx) < 0) {
        perror("Error setting up io context");
        exit (-1);
    }
    if (io_set_pagefault_context(ioctx) < 0) {
        perror("Cannot set pagefault context");
        exit(-1);
    }

    srand(time(0));
    
    /* Create all workers */
    nr_worker = argc - 1;
    for (i = 0; i < nr_worker; i++) {
        struct worker_context *worker_context;
        int fd;

        worker_context = malloc(sizeof(struct worker_context));
        if (!worker_context) {
            perror("Error allocating worker context");
            exit(-1);
        }
        memset(worker_context, 0, sizeof(struct worker_context));
        co_workers[i] = co_create(worker, worker_context->worker_stack, STACK_SIZE);
        co_workers[i]->user = worker_context;
    
        worker_context->id = i;

        /* Open and mmap the file */
        fd = open(argv[i+1], O_RDONLY);
        if (fd < 0) {
            printf("Error opening file %s\n", argv[i+1]);
            perror("");
            exit(-1);
        }
        if (fstat(fd, &st)) {
            printf("Cannot stat file %s\n", argv[i+1]);
            perror("");
            exit(-1);
        }
        worker_context->buffer = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
        worker_context->length = st.st_size;

        if (worker_context->buffer == MAP_FAILED) {
            printf ("Error mmapping %s\n", argv[i+1]);
            perror ("");
            exit(-1);
        }
    }

    /* Create the scheduler */
    co_scheduler = co_create (scheduler, &scheduler_stack, STACK_SIZE);
    
    /* Create the yielding function ucontext */
    getcontext(&uctx_yield);
    uctx_yield.uc_stack.ss_sp = yield_function_stack;
    uctx_yield.uc_stack.ss_size = STACK_SIZE;
    makecontext(&uctx_yield, pagefault_yield, 0);

    swapcontext(&uctx_main, &uctx_yield);
    
    /* Start the crank */
    co_call (co_scheduler, 0);

    return 0;   /* never reached */
}
